#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# warden-map.py
#
# Copyright (C) 2016 Cesnet z.s.p.o
# Use of this source is governed by a 3-clause BSD-style license, see LICENSE file.

import json
import codecs
import time
import argparse
import GeoIP
import requests

def getLastEvents(client, key, cert):

  res = requests.post(
      'https://warden-hub.cesnet.cz/warden3/getEvents?client=%s' % (client,),
      cert=(cert, key)
  )

  data = res.json()
  i = 0
  eventsList = []
  for p in data['events']:
    event = {}
    for key, value in { 'event': 'Category', 'time': 'DetectTime', 'origin': 'Source', 'destination': 'Target'}.items():
      if value in p:
        if (key == 'origin') or (key == 'destination'):
          event[key] = {}
          if 'IP4' in p[value][0]:
            event[key]['ip'] = p[value][0]['IP4'][0]
          else:
            event[key] = {}
        elif (key == 'event'):
          event[key] = ', '.join(p[value])
        else:
          event[key] = p[value]
      else:
        if (key == 'origin') or (key == 'destination'):
          event[key] = {}
        else:
          event[key] = {}
    if 'ip' in event['origin']:
      eventsList.append(event)
      i += 1

  return eventsList

def getGeolocation(ip, db):

  data = db.record_by_addr(ip)

  if not data:
    return {}
  else:
    return {
      'latitude': data['latitude'], 
      'longitude': data['longitude'], 
      'country_name': data['country_name'] if data['country_name'] else None, 
      'city': data['city'] if data['city'] else None
    }

def main(args):

  client = args.client[0]
  key    = args.key[0]
  cert   = args.cert[0]

  if args.output is not None:
    path = args.output[0] + 'warden-map.json'
  else:
    path = 'warden-map.json'

  db = GeoIP.open("GeoLiteCity.dat", GeoIP.GEOIP_MEMORY_CACHE)
  db.set_charset(GeoIP.GEOIP_CHARSET_UTF8)

  wardenEvents = getLastEvents(client, key, cert)

  for p in wardenEvents:
    for target in {'origin', 'destination'}:
      geoData = {}
      if 'ip' in p[target]:
        geoData = getGeolocation(p[target]['ip'], db)
        for value in {'latitude', 'longitude', 'country_name', 'city'}:
          if value in geoData:
            if not geoData[value]:
              p[target][value] = "???"
            else:
              p[target][value] = geoData[value]
          else:
            p[target][value] = "???"

      else:
        p[target]['ip'] = "???"
        p[target]['country_name'] = "Czech Republic"
        p[target]['city'] = "???"
        p[target]['latitude'] = 49.743
        p[target]['longitude'] = 15.338

  wardenEvents.append(int(time.time()));

  with open(path, 'w') as outfile:
    json.dump(wardenEvents, outfile)


if __name__ == '__main__':
  parser = argparse.ArgumentParser(description='Creates warden-map.json for warden-map.html frontend.',
                                  formatter_class=lambda prog: argparse.HelpFormatter(prog,max_help_position=30))

  parser.add_argument('--output', metavar='path/', type=str,
                    nargs=1, help='path where warden-map.json should be saved')

  requiredNamed = parser.add_argument_group('required arguments')

  requiredNamed.add_argument('--client', metavar='<org.ex.cl>', type=str, required=True,
                    nargs=1, help='client name')
  requiredNamed.add_argument('--key', metavar='path/key.pem', type=str, required=True,
                    nargs=1, help='SSL key for a client')
  requiredNamed.add_argument('--cert', metavar='path/cert.pem', type=str, required=True,
                    nargs=1, help='SSL cert for a client')

  args = parser.parse_args()
  main(args)
